import pytest
import json
from rdv.schema import Schema
from rdv.feature import FloatFeature
from rdv.stats import NumericStats, CategoricStats
from rdv.extractors.vision.similarity import FixedSubpatchSimilarity


def test_schema_jcr():
    extractor = FixedSubpatchSimilarity(
        patch={"x0": 0, "y0": 0, "x1": 64, "y1": 64}, refs=["adf8d224cb8786cc"], nrefs=1
    )
    stats = NumericStats(min=0, max=1, mean=0.5, std=0.02, percentiles=range(101))
    component = FloatFeature(name="testcomponent", extractor=extractor, stats=stats)
    component2 = FloatFeature(name="testcomponent2", extractor=extractor, stats=stats)

    schema = Schema(name="Testing", version="1.0.0", features=[component, component2])
    schema_jcr = schema.to_jcr()
    assert len(schema_jcr["features"]) == 2
    jsonstr = json.dumps(schema_jcr)  # Should not throw an error
    assert len(jsonstr) > 0
    schema_restored = Schema.from_jcr(schema_jcr)

    assert schema.name == schema_restored.name
    assert schema.version == schema_restored.version
    assert all([c1 == c2 for (c1, c2) in zip(schema.features.keys(), schema_restored.features.keys())])



import numpy as np
import torch
import torchvision.models as models
from torch.utils.data import DataLoader, Dataset
from torchvision import transforms
from PIL import Image

from rdv.extractors.structured.kmeans import KMeansOutlierScorer

# Loosely based on "Deep Nearest Neighbor Anomaly Detection": https://arxiv.org/abs/2002.10445


class ImageDataset(Dataset):
    """Face Landmarks dataset."""

    def __init__(self, loaded_data, transform=None):

        self.loaded_data = loaded_data
        self.transform = transform

    def __len__(self):
        return len(self.loaded_data)

    def __getitem__(self, idx):
        sample = self.loaded_data[idx]
        if self.transform:
            sample = self.transform(sample)
        return sample


class DN2OutlierScorer(KMeansOutlierScorer):
    def __init__(self, k=16, clusters=None, dist="euclidean"):
        super().__init__(k=k, clusters=clusters, dist=dist)
        self.mobilenet = models.mobilenet_v2(pretrained=True).eval()
        self.tfs = transforms.Compose(
            [transforms.ToTensor(), transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])]
        )

    def extract_feature(self, data):
        if not isinstance(data, Image.Image):
            raise ValueError(f"data must be of type PIL.Image.Image, not {data.shape}")
        batchtf = self.tfs(data)[None, :]
        feats = self.mobilenet(batchtf).detach().numpy()
        return super().extract_feature(data=feats)

    def build(self, data, batch_size=16):
        dataset = ImageDataset(loaded_data=data, transform=self.tfs)
        # data is a list of images here
        dataloader = DataLoader(dataset, shuffle=False, batch_size=batch_size, drop_last=True)
        embeddings = torch.zeros(size=(len(data), 1000))
        for batchidx, batchtf in enumerate(dataloader):
            # batchtf = self.tfs(batch)
            features = self.mobilenet(batchtf).detach()  # .numpy()
            startidx = batchidx * 16
            stopidx = startidx + len(features)
            embeddings[startidx:stopidx, :] = features

        embeddings = embeddings[embeddings.abs().sum(axis=1) != 0]
        embeddings = embeddings.numpy()
        super().build(data=embeddings)



from PIL import Image
import numpy as np


from rdv.extractors import FeatureExtractor
from rdv.globals import DataException
from rdv.feature import IntFeature, FloatFeature, CategoricFeature


class ElementExtractor(FeatureExtractor):
    """
    Extract one element from a vector
    """

    def __init__(self, element):
        self.element = element

    """ELEMENT"""

    @property
    def element(self):
        return self._element

    @element.setter
    def element(self, value):
        if not (isinstance(value, str) or isinstance(value, int)):
            raise DataException("element ot extract must be int or str")
        self._element = value

    def extract_feature(self, data):
        return data[self.element]

    """Serializable interface """

    def to_jcr(self):
        data = {
            "element": self.element,
        }
        return data

    @classmethod
    def from_jcr(cls, jcr):
        return cls(**jcr)

    """Buildable interface"""

    def build(self, data):
        pass

    def is_built(self):
        return True

    def __str__(self):
        return f"{self.__class__.__name__}(element={self.element})"


def construct_features(dtypes):
    components = []
    for key in dtypes.index:
        # Check type: Numeric or categoric
        extractor = ElementExtractor(element=key)
        if np.issubdtype(dtypes[key], np.floating):
            component = FloatFeature(name=key, extractor=extractor)
        elif np.issubdtype(dtypes[key], np.integer):
            component = IntFeature(name=key, extractor=extractor)
        elif dtypes[key] == np.dtype("O"):
            component = CategoricFeature(name=key, extractor=extractor)
        else:
            raise ValueError(f"dtype {dtypes[key]} not supported.")

        components.append(component)

    return components



import json

import tempfile
import time
import webbrowser
from multiprocessing import Process

from abc import ABC, abstractmethod


class SchemaStateException(Exception):
    pass


class DataException(Exception):
    pass


class Serializable(ABC):
    def class2str(self):
        module = str(self.__class__.__module__)
        classname = str(self.__class__.__name__)
        return f"{module}.{classname}"

    @abstractmethod
    def to_jcr(self):
        """Return a JSON compatible representation of the object. Will generally return a dict cintaining the objects state, but can return anything JSON serializable. json.dumps(xxx) will be called on the output xxx of this function."""
        # Return a json-compatible representation
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def from_jcr(cls, jcr):
        """Given the JSON compatible representation from the function above, load an object of this type with the desired state.

        Parameters
        ----------
        jcr : [dict]
            The jcr representation returned from the `to_jcr` function above. Will generally be a dict, but can be anything JSON serializable.

        Returns
        -------
        obj : this
            An object with type of your extractor.
        """
        # Load a json-compatible representation
        raise NotImplementedError()


class Buildable(ABC):
    @abstractmethod
    def build(self, data):
        """Your feature extractor must be Buildable. This means that it may use data to set some reference values, used to calculate the feature to be extracted from a data sample. A good example for this is the `rdv.extractors.structured.KMeansOutlierScorer` extractor, which clusters the data at building time and saves those clusters as reference in the objects state. If you dont require and buildabe state, like the `rdv.extractors.structured.ElementExtractor`, don't do anything in this function.

        Parameters
        ----------
        data : any
            The set of data available at building time. Can be any type you want.
        Returns
        -------
        None
        """
        raise NotImplementedError

    @abstractmethod
    def is_built(self):
        """
        Check whether the object has been built. Typically, this method checks whether the required references for the object is set. If your FeatureExtractor does not use any references, simply return True.

        Returns
        -------
        is_built : bool
        """
        raise NotImplementedError


class Configurable(ABC):
    @classmethod
    @abstractmethod
    def configure_interactive(self, loaded_data, mode="inline"):
        raise NotImplementedError



from abc import ABC, abstractmethod

from rdv.globals import (
    Buildable,
    Serializable,
)


class FeatureExtractor(Serializable, Buildable, ABC):
    @abstractmethod
    def extract_feature(self, data):
        """Extracts a feature from a data instance.

        Parameters
        ----------
        data : any
            The data instance you want to extract a feature from. The type is up to you.

        """
        raise NotImplementedError

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return str(self)


class NoneExtractor(FeatureExtractor):
    def to_jcr(self):
        data = {}
        return data

    @classmethod
    def from_jcr(cls, jcr):
        return cls()

    def extract(self, data):
        """Extracts a feature from the data.

        Args:
            data (Object): [description]
        """
        return 0

    def build(self, data):
        pass

    def is_built(self):
        return True

    def extract_feature(self, data):
        pass



from collections.abc import Iterable
from pydoc import locate
from abc import ABC, abstractmethod

import numpy as np
import pandas as pd
from plotly import graph_objects as go
import plotly.express as px
from plotly.subplots import make_subplots

from rdv.globals import (
    Buildable,
    Configurable,
    Serializable,
    DataException,
    SchemaStateException,
)
from rdv.stats import CategoricStats, NumericStats, equalize_domains
from rdv.tags import Tag, SCHEMA_ERROR, SCHEMA_FEATURE
from rdv.extractors import NoneExtractor

PLOTLY_COLORS = px.colors.qualitative.Plotly
HIST_N_SAMPLES = 1000


class Feature(Serializable, Buildable, ABC):
    def __init__(self, name="default_name", extractor=None):
        self.name = str(name)
        if extractor is None:
            self.extractor = NoneExtractor()
        else:
            self.extractor = extractor
        if isinstance(self.extractor, Configurable):
            self.extractor.idfr = self.name

        self.stats = None

    """Serializable interface """

    def to_jcr(self):
        data = {
            "name": self.name,
            "extractor_class": self.extractor.class2str(),
            "extractor_state": self.extractor.to_jcr(),
            "stats": self.stats.to_jcr(),
        }
        return data

    def build_extractor(self, loaded_data):
        self.extractor.build(loaded_data)

    def build_stats(self, loaded_data):
        print(f"Compiling stats for {self.name}")
        features = self.extract_features(loaded_data)
        self.stats.build(features)

    def extract_features(self, loaded_data):
        features = []
        if isinstance(loaded_data, pd.DataFrame):
            features = self.extractor.extract_feature(loaded_data)
        elif isinstance(loaded_data, Iterable):
            for data in loaded_data:
                features.append(self.extractor.extract_feature(data))
        else:
            raise DataException("loaded_data should be a DataFrame or Iterable")
        return features

    def build(self, data):
        # Compile extractor
        self.build_extractor(data)
        # Configure stats
        self.build_stats(data)

    def is_built(self):
        return self.extractor.is_built() and self.stats.is_built()

    def requires_config(self):
        return isinstance(self.extractor, Configurable) and not self.extractor.is_configured()

    def check(self, data):

        feature = self.extractor.extract_feature(data)
        # Make a tag from the feature
        feat_tag = self.feature2tag(feature)
        # Check min, max, nan or None and raise data error
        err_tag = self.check_invalid(feature)
        tags = [feat_tag, err_tag]
        # Filter Nones
        tags = [tag for tag in tags if tag is not None]
        return tags

    @abstractmethod
    def feature2tag(self, feature):
        pass

    @abstractmethod
    def check_invalid(self, feature):
        pass

    def __repr__(self):
        return str(self)

    @abstractmethod
    def plot(self, poi, size, secondary):
        pass

    @abstractmethod
    def compare(self, other, size=(800, 500)):
        pass

    def layout_settings(self, size):
        return dict(
            margin=dict(l=0, r=0, t=50, b=0),
            height=size[1],
            width=size[0],
            template="plotly_white",
            # title_text=f"cdf for {self.name}",
            xaxis_title=self.name,
            legend=dict(orientation="h", yanchor="bottom", y=1, xanchor="left", x=0),
        )


class FloatFeature(Feature):
    def __init__(self, name="default_name", extractor=None, stats=None):
        super().__init__(name=name, extractor=extractor)
        self._stats = None
        self.stats = stats

    """
    PROPERTIES
    """

    @property
    def stats(self):
        return self._stats

    @stats.setter
    def stats(self, value):
        if value is None:
            self._stats = NumericStats()
        elif isinstance(value, NumericStats):
            self._stats = value
        else:
            raise DataException(f"stats for a NumericComponant should be of type NumericStats, not {type(value)}")

    def feature2tag(self, feature):
        if not np.isnan(feature):
            return Tag(name=self.name, value=float(feature), type=SCHEMA_FEATURE)
        else:
            return None

    def check_invalid(self, feature):
        tagname = f"{self.name}-err"
        if feature is None:
            return Tag(name=tagname, value="Value None", type=SCHEMA_ERROR)
        elif np.isnan(feature):
            return Tag(name=tagname, value="Value NaN", type=SCHEMA_ERROR)
        elif feature > self.stats.max:
            return Tag(name=tagname, value="Value > max", type=SCHEMA_ERROR)
        elif feature < self.stats.min:
            return Tag(name=tagname, value="Value < min", type=SCHEMA_ERROR)
        else:
            return None

    @classmethod
    def from_jcr(cls, jcr):
        classpath = jcr["extractor_class"]
        extr_class = locate(classpath)
        if extr_class is None:
            NameError(f"Could not locate {classpath}")

        extractor = extr_class.from_jcr(jcr["extractor_state"])
        stats = NumericStats.from_jcr(jcr["stats"])
        name = jcr["name"]
        return cls(name=name, extractor=extractor, stats=stats)

    def __str__(self):
        return f"FloatFeature(name={self.name}, extractor={self.extractor})"

    def plot(self, poi=None, size=(800, 500), secondary=True):
        if not self.is_built():
            raise SchemaStateException(f"Feature {self.name} has not been built. Cannot plot unbuilt components.")
        fig = go.Figure()

        # fig = go.Figure()
        fig.add_trace(plot_cdf(self.stats.percentiles))
        if secondary:
            fig.add_trace(plot_histogram(self.stats.sample(n=HIST_N_SAMPLES)))

        fig.update_layout(**self.layout_settings(size))
        range_min, range_max = self.stats.percentiles[0], self.stats.percentiles[-1]
        print(f"POI: {poi}")
        if poi and (isinstance(poi, float) or isinstance(poi, int)):
            fig.add_shape(
                type="line",
                x0=poi,
                y0=0,
                x1=poi,
                y1=100,
                line=dict(color=PLOTLY_COLORS[-1], width=1),
                name="poi",
            )
            if poi < range_min:
                range_min = poi
            if poi > range_max:
                range_max = poi

        fig.update_xaxes(range=[range_min, range_max])

        return fig

    def compare(self, other, size=(800, 500)):
        fig = go.Figure()
        fig.add_trace(
            go.Scatter(
                x=self.stats.percentiles,
                y=list(
                    range(
                        len(
                            self.stats.percentiles,
                        )
                    )
                ),
                mode="lines",
                name="self",
            )
        )
        fig.add_trace(
            go.Scatter(
                x=other.stats.percentiles,
                y=list(
                    range(
                        len(
                            other.stats.percentiles,
                        )
                    )
                ),
                mode="lines",
                name="other",
            )
        )
        fig.update_layout(**self.layout_settings(size))

        return fig


class IntFeature(Feature):
    def __init__(self, name="default_name", extractor=None, stats=None):
        super().__init__(name=name, extractor=extractor)
        self._stats = None
        self.stats = stats

    """
    PROPERTIES
    """

    @property
    def stats(self):
        return self._stats

    @stats.setter
    def stats(self, value):
        if value is None:
            self._stats = NumericStats()
        elif isinstance(value, NumericStats):
            self._stats = value
        else:
            raise DataException(f"stats for a NumericComponant should be of type NumericStats, not {type(value)}")

    def feature2tag(self, feature):
        if not np.isnan(feature):
            return Tag(name=self.name, value=float(feature), type=SCHEMA_FEATURE)
        else:
            return None

    def check_invalid(self, feature):
        tagname = f"{self.name}-err"
        if feature is None:
            return Tag(name=tagname, value="Value None", type=SCHEMA_ERROR)
        elif np.isnan(feature):
            return Tag(name=tagname, value="Value NaN", type=SCHEMA_ERROR)
        elif feature > self.stats.max:
            return Tag(name=tagname, value="Value > max", type=SCHEMA_ERROR)
        elif feature < self.stats.min:
            return Tag(name=tagname, value="Value < min", type=SCHEMA_ERROR)
        else:
            return None

    @classmethod
    def from_jcr(cls, jcr):
        classpath = jcr["extractor_class"]
        extr_class = locate(classpath)
        if extr_class is None:
            NameError(f"Could not locate {classpath}")

        extractor = extr_class.from_jcr(jcr["extractor_state"])
        stats = NumericStats.from_jcr(jcr["stats"])
        name = jcr["name"]
        return cls(name=name, extractor=extractor, stats=stats)

    def __str__(self):
        return f"IntFeature(name={self.name}, extractor={self.extractor})"

    def plot(self, poi=None, size=(800, 500), secondary=True):
        if not self.is_built():
            raise SchemaStateException(f"Feature {self.name} has not been built. Cannot plot unbuilt components.")
        fig = go.Figure()

        # fig = go.Figure()
        fig.add_trace(plot_cdf(self.stats.percentiles))
        if secondary:
            fig.add_trace(plot_histogram(self.stats.sample(n=HIST_N_SAMPLES)))

        fig.update_layout(**self.layout_settings(size))
        range_min, range_max = self.stats.percentiles[0], self.stats.percentiles[-1]
        print(f"POI: {poi}")
        if poi and (isinstance(poi, float) or isinstance(poi, int)):
            fig.add_shape(
                type="line",
                x0=poi,
                y0=0,
                x1=poi,
                y1=100,
                line=dict(color=PLOTLY_COLORS[-1], width=1),
                name="poi",
            )
            if poi < range_min:
                range_min = poi
            if poi > range_max:
                range_max = poi

        fig.update_xaxes(range=[range_min, range_max])

        return fig

    def compare(self, other, size=(800, 500)):
        fig = go.Figure()
        fig.add_trace(
            go.Scatter(
                x=self.stats.percentiles,
                y=list(
                    range(
                        len(
                            self.stats.percentiles,
                        )
                    )
                ),
                mode="lines",
                name=f"self",
            )
        )
        fig.add_trace(
            go.Scatter(
                x=other.stats.percentiles,
                y=list(
                    range(
                        len(
                            other.stats.percentiles,
                        )
                    )
                ),
                mode="lines",
                name=f"other",
            )
        )
        fig.update_layout(**self.layout_settings(size))

        return fig


class CategoricFeature(Feature):

    # Domain, domain distribution
    def __init__(self, name="default_name", extractor=None, stats=None):
        super().__init__(name=name, extractor=extractor)
        self._stats = None
        self.stats = stats

    """
    PROPERTIES
    """

    @property
    def stats(self):
        return self._stats

    @stats.setter
    def stats(self, value):
        if value is None:
            self._stats = CategoricStats()
        elif isinstance(value, CategoricStats):
            self._stats = value
        else:
            raise DataException(f"stats for a NumericComponant should be of type CategoricStats, not {type(value)}")

    def feature2tag(self, feature):
        if isinstance(feature, str) or not np.isnan(feature):
            return Tag(name=self.name, value=feature, type=SCHEMA_FEATURE)
        else:
            return None

    def check_invalid(self, feature):
        tagname = f"{self.name}-err"
        if feature is None:
            return Tag(name=tagname, value="Value None", type=SCHEMA_ERROR)
        elif pd.isnull(feature):
            return Tag(name=tagname, value="Value NaN", type=SCHEMA_ERROR)
        elif feature not in self.stats.domain_counts:
            return Tag(name=tagname, value="Domain Error", type=SCHEMA_ERROR)
        else:
            return None

    @classmethod
    def from_jcr(cls, jcr):
        classpath = jcr["extractor_class"]
        extr_class = locate(classpath)
        if extr_class is None:
            NameError(f"Could not locate {classpath}")

        extractor = extr_class.from_jcr(jcr["extractor_state"])
        stats = CategoricStats.from_jcr(jcr["stats"])
        name = jcr["name"]

        return cls(name=name, extractor=extractor, stats=stats)

    def __str__(self):
        return f"CategoricFeature(name={self.name}, extractor={self.extractor})"

    def plot(self, poi=None, size=(800, 500), secondary=True):
        if not self.is_built():
            raise SchemaStateException(f"Feature {self.name} has not been built. Cannot plot unbuilt components.")

        domain = sorted(list(self.stats.domain_counts.keys()))
        # Let's be absolutely sure the domain is always in the same order
        p = [self.stats.domain_counts[k] * 100 for k in domain]
        cdf = np.cumsum(p).astype("int")
        fig = go.Figure()
        fig.add_trace(go.Bar(x=domain, y=p, name="value freq"))
        if secondary:
            fig.add_trace(go.Scatter(x=domain, y=cdf, name="cdf"))

        # fig.update_xaxes(range=[self.stats.percentiles[0], self.stats.percentiles[-1]])
        fig.update_layout(**self.layout_settings(size))

        if poi and isinstance(poi, str):
            fig.add_shape(
                type="line",
                x0=poi,
                y0=0,
                x1=poi,
                y1=100,
                line=dict(color=PLOTLY_COLORS[-1], width=1),
                name="poi",
            )
        return fig

    def compare(self, other, size=(800, 500)):
        def get_bartrace(domain_counts, name):
            return go.Bar(x=list(domain_counts.keys()), y=list(domain_counts.values()), name=name)

        domain_f1, domain_f2, _ = equalize_domains(a=self.stats.domain_counts, b=other.stats.domain_counts)

        fig = go.Figure()
        fig.add_trace(get_bartrace(domain_f1, name="self"))
        fig.add_trace(get_bartrace(domain_f2, name="other"))
        # ddof = (len(domain) - 1) ** 2
        # stat, pvalue = self.stats.test_drift(other.stats)
        # If pvalue small enough -> H0 does not hold -> drift
        fig.update_layout(**self.layout_settings(size))

        return fig


def plot_histogram(samples, range=None, dtype="float"):
    # px = sample_cdf(percentiles=percentiles, n_samples=1000, dtype=dtype)
    hist, edges = np.histogram(samples, bins=100, range=range)
    hist = hist / hist.sum() * 100
    widths = np.diff(edges)
    bin_centers = (edges[:-1] + edges[1:]) / 2

    return go.Bar(x=bin_centers, y=hist, width=widths, name="histogram")


def plot_cdf(percentiles):
    return go.Scatter(x=percentiles, y=list(range(len(percentiles))), name="cdf")



from PIL import Image, ImageFilter
import numpy as np

from rdv.extractors import FeatureExtractor


class AvgIntensity(FeatureExtractor):

    _config_attrs = []
    _compile_attrs = []
    _ccable_deps = []
    _attrs = _config_attrs + _compile_attrs + _ccable_deps

    def __init__(self):
        pass

    def extract_feature(self, data):
        img = data
        img = img.convert("L")
        return float(np.array(img).mean())

    """Serializable inteface """

    def to_jcr(self):
        return {}

    @classmethod
    def from_jcr(cls, jcr):
        return cls()

    """Buildable interface"""

    def build(self, data):
        pass

    def is_built(self):
        return True



import pytest
import json
from rdv.schema import Schema
from rdv.feature import FloatFeature
from rdv.stats import NumericStats, CategoricStats
from rdv.extractors.vision.similarity import FixedSubpatchSimilarity


def test_stats_none():
    stats = NumericStats()
    jcr = stats.to_jcr()
    for attr in NumericStats._attrs:
        assert jcr[attr] is None


def test_stats_partial_none():
    params = dict(min=0, max=1, mean=0.5, std=0.02, pinv=0.1, percentiles=list(range(101)))
    stats = NumericStats(**params)
    jcr = stats.to_jcr()
    for attr in NumericStats._attrs:
        assert jcr[attr] == params[attr]


def test_num_stats_build():
    stats = NumericStats()
    assert not stats.is_built()
    stats = NumericStats(
        min=0,
        max=1,
        mean=0.5,
        std=0.02,
    )
    assert not stats.is_built()
    stats = NumericStats(min=0, max=1, mean=0.5, percentiles=range(101), pinv=0)
    assert not stats.is_built()
    stats = NumericStats(min=0, max=1, std=0.02, percentiles=range(101), pinv=0)
    assert not stats.is_built()
    stats = NumericStats(min=0, mean=0.5, std=0.02, percentiles=range(101), pinv=0)
    assert not stats.is_built()
    stats = NumericStats(max=1, mean=0.5, std=0.02, percentiles=range(101), pinv=0)
    assert not stats.is_built()
    stats = NumericStats(min=0, max=1, mean=0.5, std=0.02, percentiles=range(101))
    assert not stats.is_built()
    stats = NumericStats(min=0, max=1, mean=0.5, std=0.02, percentiles=range(101), pinv=0)
    assert stats.is_built()


def test_cat_stats_build():
    stats = CategoricStats()
    assert not stats.is_built()
    stats = CategoricStats(domain_counts={"a": 0.5, "b": 0.5}, pinv=0.1)
    assert not stats.is_built()
    stats = CategoricStats(domain_counts={"a": 0.5, "b": 0.5}, samplesize=10)
    assert not stats.is_built()
    stats = CategoricStats(pinv=0.1, samplesize=10)
    assert not stats.is_built()
    stats = CategoricStats(domain_counts={"a": 0.5, "b": 0.5}, pinv=0.1, samplesize=10)
    assert stats.is_built()


def test_subpatchsimilarity_extractor_buildable():
    with pytest.raises(ValueError):
        extractor = FixedSubpatchSimilarity(patch=None)

    extractor = FixedSubpatchSimilarity(patch={"x0": 0, "y0": 0, "x1": 64, "y1": 64})
    assert not extractor.is_built()

    extractor = FixedSubpatchSimilarity(
        patch={"x0": 0, "y0": 0, "x1": 64, "y1": 64}, refs=["adf8d224cb8786cc"], nrefs=1
    )
    assert extractor.is_built()


def test_feature_buildable():
    component = FloatFeature(name="testcomponent")
    assert not component.is_built()

    extractor = FixedSubpatchSimilarity(
        patch={"x0": 0, "y0": 0, "x1": 64, "y1": 64}, refs=["adf8d224cb8786cc"], nrefs=1
    )
    stats = NumericStats(min=0, max=1)
    component = FloatFeature(name="testcomponent", extractor=extractor, stats=stats)
    assert not stats.is_built()
    assert extractor.is_built()
    assert not component.is_built()

    extractor = FixedSubpatchSimilarity(
        patch={"x0": 0, "y0": 0, "x1": 64, "y1": 64}, refs=["adf8d224cb8786cc"], nrefs=1
    )
    stats = NumericStats(min=0, max=1, mean=0.5, std=0.02, percentiles=range(101), pinv=0)
    component = FloatFeature(name="testcomponent", extractor=extractor, stats=stats)
    assert stats.is_built()
    assert extractor.is_built()
    assert component.is_built()


def test_schema_buildable():
    component = FloatFeature(name="testcomponent")
    schema = Schema(name="Testing", version="1.0.0", features=[component])
    assert not schema.is_built()

    extractor = FixedSubpatchSimilarity(
        patch={"x0": 0, "y0": 0, "x1": 64, "y1": 64}, refs=["adf8d224cb8786cc"], nrefs=1
    )
    stats = NumericStats(min=0, max=1)
    component = FloatFeature(name="testcomponent", extractor=extractor, stats=stats)
    schema = Schema(name="Testing", version="1.0.0", features=[component])

    assert not schema.is_built()

    extractor = FixedSubpatchSimilarity(
        patch={"x0": 0, "y0": 0, "x1": 64, "y1": 64}, refs=["adf8d224cb8786cc"], nrefs=1
    )
    stats = NumericStats(min=0, max=1, mean=0.5, std=0.02, percentiles=range(101), pinv=0)
    component = FloatFeature(name="testcomponent", extractor=extractor, stats=stats)
    schema = Schema(name="Testing", version="1.0.0", features=[component, component])

    assert schema.is_built()



import base64
import json
import random
from pathlib import Path

import dash_core_components as dcc
import dash_html_components as html
import imagehash
import numpy as np
import plotly.graph_objects as go
import rdv.dash
from dash.dependencies import Input, Output
from rdv.dash.helpers import get_dash
from rdv.extractors import FeatureExtractor
from rdv.globals import Configurable


class FixedSubpatchSimilarity(FeatureExtractor, Configurable):

    _attrs = ["patch", "refs"]
    _patch_keys = ["x0", "y0", "x1", "y1"]

    def __init__(self, patch, refs=None, nrefs=10, idfr=None):
        """[summary]

        Args:
            patch ([int], optional): [description]. The x0, y0, x1, y1 of the patch to look at.
            refs ([np.array], optional): [description]. References of what the patch should look like
        """
        self._nrefs = None
        self._patch = None
        self._refs = None
        self._idfr = None

        self.patch = patch
        self.nrefs = nrefs
        self.refs = refs
        self.idfr = idfr

    """
    PROPERTIES
    """

    @property
    def patch(self):
        return self._patch

    @patch.setter
    def patch(self, value):

        if isinstance(value, dict):
            self._patch = {key: value[key] for key in self._patch_keys}
        elif isinstance(value, list) and len(value) == 4:
            self._patch = {key: value[i] for i, key in enumerate(self._patch_keys)}
        else:
            raise ValueError(f"patch must be a dict or list, not {type(value)}")
        # make sure the correct keys are there
        print(f"Patch set to: {self._patch} for {self}")

    @property
    def refs(self):
        return self._refs

    @refs.setter
    def refs(self, value):
        if value is None:
            self._refs = None
            return

        if not (isinstance(value, list) and len(value) == self.nrefs):
            raise ValueError(f"refs should be a list of length {self.nrefs}")

        parsed_refs = []
        for ref in value:
            if isinstance(ref, imagehash.ImageHash):
                parsed_refs.append(ref)
            elif isinstance(ref, str):
                parsed_refs.append(imagehash.hex_to_hash(ref))
            else:
                raise ValueError(f"refs should either be str or ImageHash, not {type(ref)}")

        self._refs = parsed_refs

    @property
    def nrefs(self):
        return self._nrefs

    @nrefs.setter
    def nrefs(self, value):
        value = int(value)
        if not (isinstance(value, int) and value > 0):
            self._nrefs = None
            raise ValueError(f"nrefs should be a an int > 0")
        self._nrefs = value

    @property
    def idfr(self):
        return self._idfr

    @idfr.setter
    def idfr(self, value):
        self._idfr = str(value)

    """Feature extractor"""

    def extract_feature(self, data):
        phash = self._extract(data)
        dist = min(abs(ref - phash) for ref in self.refs)
        return dist

    def _extract(self, data):
        patch = [self.patch["x0"], self.patch["y0"], self.patch["x1"], self.patch["y1"]]
        crop = data.crop(box=patch)
        phash = imagehash.phash(crop)
        return phash

    """Serializable interface """

    def to_jcr(self):
        data = {
            "patch": self.patch,
            "refs": [str(ref) for ref in self.refs] if self.refs is not None else None,
            "nrefs": self.nrefs,
        }
        return data

    @classmethod
    def from_jcr(cls, jcr):
        patch, refs, nfres, idfr = None, None, None, None
        if "patch" in jcr:
            patch = jcr["patch"]
        if "nrefs" in jcr:
            nrefs = jcr["nrefs"]
        if "refs" in jcr:
            refs = jcr["refs"]
        if "idfr" in jcr:
            refs = jcr["idfr"]

        return cls(patch=patch, refs=refs, nrefs=nrefs, idfr=idfr)

    """Buildable interface"""

    def build(self, data):
        refs = []
        chosen_samples = random.choices(data, k=self.nrefs)
        for sample in chosen_samples:
            ref = self._extract(sample)
            refs.append(ref)
        self.refs = refs

    def is_built(self):
        return self.refs is not None and len(self.refs) == self.nrefs

    def __str__(self):
        return f"{self.class2str()} ({self.idfr})"

    """Configurable interface"""

    def set_config(self, data):
        self.patch = data

    def is_configured(self):
        return isinstance(self.patch, dict) and all(key in self.patch for key in self._patch_keys)

    @classmethod
    def configure_interactive(cls, loaded_data, mode="inline"):
        def render_controls():
            return html.Div(
                [
                    html.H3("Controls", className="Subhead"),
                    html.Form(
                        html.Div(
                            className="form",
                            children=[
                                html.Div(
                                    id="page-controls",
                                    children=[],
                                ),
                                html.Div(
                                    className="form-group-header",
                                    children=html.Label("Patch Location:", className="m-2"),
                                ),
                                html.Div(
                                    className="form-group-body",
                                    children=html.Pre(
                                        id="patch-config",
                                        className="m-2 similarity-code-view",
                                    ),
                                ),
                                html.Div(
                                    className="form-group-body",
                                    children=html.Div(
                                        id="extractor-constructor",
                                        className="m-2 codelike-content",
                                    ),
                                ),
                            ],
                        ),
                    ),
                ],
                className="col-4 float-left",
                id="left-column-div",
            )

        def render_img_view():
            return html.Div(
                [
                    html.H3("View", className="Subhead"),
                    dcc.Graph(
                        id="graph-image",
                    ),
                    html.Div(id="hidden-dummy"),
                ],
                className="col-8 float-left",
            )

        def render_figure(active_img_idx, patch, editable=True):
            active_img = loaded_data[active_img_idx].copy()
            img_width, img_height = active_img.size

            fig = go.Figure(go.Image(z=np.array(active_img)))
            fig.add_shape(
                editable=editable,
                xref="x",
                yref="y",
                layer="above",
                line={"color": "cyan"},
                opacity=0.7,
                fillcolor="cyan",
                type="rect",
                **patch,
            )
            fig.update_xaxes(showgrid=False, range=(0, img_width))
            fig.update_yaxes(showgrid=False, scaleanchor="x", range=(img_height, 0))
            fig.update_layout(autosize=False, width=800, height=800)

            return fig

        applauncher, kwargs = get_dash(mode=mode)
        app = applauncher(  # dash.Dash(
            "Data Schema Config",
            assets_folder=str((Path(rdv.dash.__file__) / "../assets").resolve()),
        )
        app.layout = html.Div(
            [
                dcc.Location(id="url", refresh=False),
                html.Div(
                    [
                        html.H1(f"Configuring extractor {cls.__name__}", className="pagehead mb-5"),
                        html.Div(
                            [
                                render_controls(),
                                render_img_view(),
                            ],
                            className="clearfix",
                        ),
                    ],
                    className="gutter-md-spacious",
                ),
            ],
            id="page-body",
            className="schema-container",
        )
        app.title = f"Configuring component {cls.__name__}"

        @app.callback(
            [
                Output(component_id="patch-config", component_property="children"),
                Output(component_id="extractor-constructor", component_property="children"),
            ],
            [Input(component_id="graph-image", component_property="relayoutData")],
        )
        def update_patch_data(shape_data):
            keys = ["x0", "y0", "x1", "y1"]

            if shape_data is None or not "shapes[0].x0" in shape_data:
                # raise PreventUpdate()
                data = {"x0": 0, "y0": 0, "x1": 64, "y1": 64}
            else:
                data = {key: int(shape_data[f"shapes[0].{key}"]) for key in keys}

            return f"{json.dumps(data, indent=4)}", f"{cls.__name__}(patch={[data[k] for k in keys]})"

        def parse_imgidx(pathname):
            if pathname == "/":
                img_idx = 0
            else:
                img_idx = int(pathname.split("/")[-1])
            return img_idx

        @app.callback(
            Output("graph-image", "figure"),
            [Input("url", "pathname"), Input("url", "search")],
            # [State("patch-config", "children")],
        )
        def render_img(pathname, search):

            if len(search) == 0:
                patch = {"x0": 0, "y0": 0, "x1": 64, "y1": 64}
            else:
                patchenc = search.split("patch=")[1]
                patch = json.loads(base64.b64decode(patchenc.encode("ascii")).decode("ascii"))

            img_idx = parse_imgidx(pathname)
            return render_figure(active_img_idx=img_idx, patch=patch)

        @app.callback(
            Output("page-controls", "children"),
            [Input("url", "pathname"), Input("patch-config", "children")],
        )
        def update_controls(pathname, patch):
            img_idx = parse_imgidx(pathname)
            if patch is None:
                patch = json.dumps(None)
            patchenc = base64.b64encode(patch.encode("ascii")).decode("ascii")

            return [
                html.Button(dcc.Link("Prev", href=f"/{img_idx-1}?patch={patchenc}"), className="btn btn-sm m-2"),
                f"Image: {img_idx} / {len(loaded_data)}",
                html.Button(dcc.Link("Next", href=f"/{img_idx+1}?patch={patchenc}"), className="btn btn-sm m-2"),
            ]

        app.run_server(debug=True, **kwargs)



from enum import Enum

from rdv.globals import Serializable


SCHEMA_ERROR = "schema-error"
SCHEMA_FEATURE = "schema-feature"
SCHEMA_FEATURE_LH = "schema-feature-lh"
SCHEMA_GLOBAL_LH = "schema-global-lh"
LABEL = "label"
METRIC = "metric"
VECTOR = "vector"
ERROR = "error"


class Tag(Serializable):
    def __init__(self, name, value, type, group=None):
        self.name = name
        self.value = value
        self.type = type
        self.group = group

    def to_jcr(self):
        jcr = {
            "type": self.type,
            "name": self.name,
            "value": self.value,
            "group": self.group,
        }
        return jcr

    @classmethod
    def from_jcr(cls, jcr):
        return cls(**jcr)

    def __str__(self):
        return f"'{self.name}:{self.value}"

    def __repr__(self):
        return f"Tag(name='{self.name}, value={self.value}, type={self.type}, group={self.group}"



import json
from pydoc import locate
from pathlib import Path

import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output

import rdv
from rdv.globals import Buildable, SchemaStateException, Serializable
from rdv.dash.helpers import get_dash
from rdv.tags import SCHEMA_FEATURE
from rdv.feature import Feature


class Schema(Serializable, Buildable):
    _attrs = ["name", "version", "features"]

    def __init__(self, name="default", version="0.0.0", features=[]):

        self._name = None
        self._version = None
        self._features = {}

        self.name = str(name)
        self.version = str(version)
        self.features = features

    """Serializable interface"""

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, str):
            raise ValueError(f"Schema name should be a string")
        self._name = value

    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, value):
        if not isinstance(value, str):
            raise ValueError(f"Schema version should be a string")
        self._version = value

    @property
    def features(self):
        return self._features

    @features.setter
    def features(self, value):
        if isinstance(value, list) and all(isinstance(feat, Feature) for feat in value):
            # Convert to dict
            self._features = {c.name: c for c in value}

        elif isinstance(value, dict) and all(isinstance(feat, Feature) for feat in value.values()):
            self._features = value
        else:
            raise ValueError(f"features must be a list[Feature] or dict[str, Feature]")

    @property
    def group_idfr(self):
        return f"{self.name}@{self.version}"

    def to_jcr(self):
        jcr = {
            "name": self.name,
            "version": self.version,
            "features": [],
        }
        features = []
        for feat in self.features.values():
            features.append({"feature_class": feat.class2str(), "feature": feat.to_jcr()})
        jcr["features"] = features
        return jcr

    @classmethod
    def from_jcr(cls, jcr):
        name = jcr["name"]
        version = jcr["version"]
        features = []
        for comp_dict in jcr["features"]:
            classpath = comp_dict["feature_class"]
            comp_jcr = comp_dict["feature"]
            compclass = locate(classpath)
            if compclass is None:
                NameError("Could not locate classpath")
            feature = compclass.from_jcr(comp_jcr)
            features.append(feature)

        return cls(name=name, version=version, features=features)

    def save(self, fpath):
        with open(fpath, "w") as f:
            json.dump(self.to_jcr(), f, indent=4)

    @classmethod
    def load(cls, fpath):
        with open(fpath, "r") as f:
            jcr = json.load(f)
        return cls.from_jcr(jcr)

    """Buildable Interface"""

    def build(self, data):

        # Build the schema
        for feat in self.features.values():
            # Compile stats
            feat.build(data)

    def is_built(self):
        return all(feat.is_built() for feat in self.features.values())

    """Other Methods"""

    def __str__(self):
        return f'Schema(name="{self.name}", version="{self.version}"'

    def set_schema_group(self, tags):
        for feature_tag in tags:
            feature_tag.group = self.group_idfr

    def check(self, data, convert_json=True):
        tags = []
        if self.is_built():
            for feature in self.features.values():
                feature_tags = feature.check(data)
                self.set_schema_group(feature_tags)
                tags.extend(feature_tags)
        else:
            raise SchemaStateException(f"Cannot check data on an unbuilt schema. Check whether all features are built.")
        if convert_json:
            tags = [t.to_jcr() for t in tags]
        return tags

    def drop_feature(self, name):
        self.features = [c for c in self.features.values() if c.name != name]

    def tags2featvalue(self, tags):
        tags_dict = {}
        for tag in tags:
            if tag["type"] == SCHEMA_FEATURE:
                tags_dict[tag["name"]] = tag["value"]
        return tags_dict

    def view(self, poi=None, mode="inline"):
        if poi is not None:
            poi_dict = self.tags2featvalue(self.check(poi))
        else:
            poi_dict = {}

        def get_feature_page(feature_name):
            return html.Div(
                id="feature-page-div",
                children=[
                    dcc.Graph(id="feature-graph", figure=self.features[feature_name].plot()),
                    dcc.Link("Back", href=f"/"),
                ],
                className="my-3",
            )

        def feature2row(feat):
            return html.Tr(
                [
                    html.Td(dcc.Link(feat.name, href=f"/{feat.name}")),
                    html.Td(feat.__class__.__name__, className="codelike-content"),
                    html.Td(
                        dcc.Graph(
                            id=f"{feat.name}-prev",
                            figure=feat.plot(size=(400, 150), poi=poi_dict.get(feat.name, None), secondary=False),
                            config={"displayModeBar": False},
                        )
                    ),
                ]
            )

        def get_feature_table():
            return html.Table(
                children=[
                    html.Thead(
                        className="schema-tablehead",
                        children=[
                            html.Tr(
                                [
                                    html.Td("Feature", className="schema-name-column"),
                                    html.Td("Type", className="schema-type-column"),
                                    html.Td("Indicator", className="schema-preview-column"),
                                ]
                            )
                        ],
                    ),
                    html.Tbody(children=[feature2row(feat) for feat in self.features.values()]),
                ]
            )

        applauncher, kwargs = get_dash(mode=mode)
        app = applauncher(  # dash.Dash(
            __name__,
            assets_folder=str((Path(rdv.dash.__file__) / "../assets").resolve()),
        )
        app.title = f"Schema viewer"
        app.layout = html.Div(
            className="schema-container",
            children=[
                dcc.Location(id="url", refresh=False),
                html.Div(
                    className="m-5",
                    children=[
                        html.H1(app.title, className="my-3"),
                        html.H3(
                            [
                                "Schema:",
                                html.Span(self.name, className="codelike-flex"),
                                ", Version:",
                                html.Span(self.version, className="codelike-flex"),
                            ],
                            className="my-3",
                        ),
                        html.Div(id="page-content"),
                    ],
                ),
            ],
        )

        @app.callback(Output("page-content", "children"), [Input("url", "pathname")])
        def display_page(pathname):
            if pathname == "/":
                return get_feature_table()
            else:
                return get_feature_page(pathname.split("/")[1])

        app.run_server(**kwargs)

    def compare(self, other, pthresh=0.05, mode="inline"):
        def drift_cell(feat_name):
            self_feat = self.features[feat_name]
            other_feat = other.features[feat_name]
            _, pvalue, drift = self_feat.stats.test_drift(other_feat.stats, pthresh=pthresh)
            if drift:
                return html.Span(className="Label mr-1 Label--red", children=f"{pvalue:.2E}")
            else:
                return html.Span(className="Label mr-1 Label--green", children=f"{pvalue:.2E}")

        def get_feature_page(feature_name):
            return html.Div(
                id="feature-page-div",
                children=[
                    dcc.Graph(
                        id="feature-graph",
                        figure=self.features[feature_name].compare(other=other.features[feature_name]),
                    ),
                    dcc.Link("Back", href=f"/"),
                ],
                className="my-3",
            )

        def feature2row(feat):
            return html.Tr(
                [
                    html.Td(dcc.Link(feat.name, href=f"/{feat.name}")),
                    html.Td(feat.__class__.__name__, className="codelike-content"),
                    html.Td(drift_cell(feat.name), className="codelike-content"),
                    html.Td(
                        dcc.Graph(
                            id=f"{feat.name}-prev",
                            figure=feat.compare(size=(400, 150), other=other.features[feat.name]),
                            config={"displayModeBar": False},
                        )
                    ),
                ]
            )

        def get_feature_table():
            return html.Table(
                children=[
                    html.Thead(
                        className="schema-tablehead",
                        children=[
                            html.Tr(
                                [
                                    html.Td("Feature", className="schema-name-column"),
                                    html.Td("Type", className="schema-type-column"),
                                    html.Td("H0 pvalue", className="schema-h0-column"),
                                    html.Td("Indicator", className="schema-preview-column"),
                                ]
                            )
                        ],
                    ),
                    html.Tbody(children=[feature2row(feat) for feat in self.features.values()]),
                ]
            )

        applauncher, kwargs = get_dash(mode=mode)
        app = applauncher(  # dash.Dash(
            __name__,
            assets_folder=str((Path(rdv.dash.__file__) / "../assets").resolve()),
        )
        app.title = f"Schema compare"
        app.layout = html.Div(
            className="schema-container",
            children=[
                dcc.Location(id="url", refresh=False),
                html.Div(
                    className="m-5",
                    children=[
                        html.H1(app.title, className="my-3"),
                        html.H3(
                            [
                                "Schema",
                                html.Span(self.name, className="codelike-flex"),
                                ", Version:",
                                html.Span(self.version, className="codelike-flex"),
                                html.Span("(self)", className="codelike-flex"),
                                " vs Schema",
                                html.Span(other.name, className="codelike-flex"),
                                ", Version:",
                                html.Span(other.version, className="codelike-flex"),
                                html.Span("(other)", className="codelike-flex"),
                            ],
                            className="my-3",
                        ),
                        html.Div(id="page-content"),
                    ],
                ),
            ],
        )

        @app.callback(Output("page-content", "children"), [Input("url", "pathname")])
        def display_page(pathname):
            if pathname == "/":
                return get_feature_table()
            else:
                return get_feature_page(pathname.split("/")[1])

        app.run_server(**kwargs)



import base64

import numpy as np

from sklearn.cluster import KMeans
from sklearn.metrics.pairwise import cosine_distances, euclidean_distances

from rdv.extractors import FeatureExtractor


class KMeansOutlierScorer(FeatureExtractor):

    dist_choices = {"euclidean": euclidean_distances, "cosine": cosine_distances}

    def __init__(self, k=16, clusters=None, dist="euclidean"):
        self._k = None
        self._clusters = None
        self._dist = None
        self.k = k
        self.clusters = clusters
        self.dist = dist

    """
    PROPERTIES
    """

    @property
    def k(self):
        return self._k

    @k.setter
    def k(self, value):
        if isinstance(value, int) and value > 0:
            self._k = value
        else:
            raise ValueError(f"k must be an int > 0, not {value} ({type(value)})")

    @property
    def clusters(self):
        return self._clusters

    @clusters.setter
    def clusters(self, value):
        if value is None:
            self._clusters = None
            return
        nclusters, dim = value.shape
        value = value.astype(np.float64)
        if isinstance(value, np.ndarray) and nclusters == self._k:
            self._clusters = value
        else:
            raise ValueError(
                f"clusters must be an np.ndarray of shape ({self.k}, dim), not {type(value)} ({value.shape})"
            )

    @property
    def dist(self):
        if self._dist in self.dist_choices:
            return self.dist_choices[self._dist]
        else:
            raise ValueError(f"Invalid distance specified: {self._dist}")

    @dist.setter
    def dist(self, value):
        if isinstance(value, str) and value in self.dist_choices:
            self._dist = value
        else:
            raise ValueError(f"dist must be str and one of {self.dist_choices.keys()}, not {value}, ({type(value)})")

    @property
    def dim(self):
        return self.clusters.shape[1]

    def extract_feature(self, data):
        def sum_2closest(distances):
            return np.sort(distances, axis=1)[:, :2].sum(axis=1)

        if data.shape == (self.dim,):
            data = data[None, :]
        elif data.shape == (1, self.dim):
            pass
        else:
            raise ValueError(f"data must be of shape {(1, self.dim)} or {(self.dim, )}, not {data.shape}")
        pairwise_dist = self.dist(data, self.clusters)
        return float(sum_2closest(pairwise_dist))

    """Buildable interface"""

    def build(self, data):
        data = np.array(data).astype(np.float64)
        km = KMeans(n_clusters=16)
        km.fit(data)
        clusters = km.cluster_centers_
        self.clusters = clusters

    def is_built(self):
        return self.clusters is not None and len(self.clusters) == self.k

    """Serializable interface"""

    def to_jcr(self):
        b64 = base64.b64encode(self.clusters).decode()
        shape = self.clusters.shape
        diststr = [k for k, v in self.dist_choices.items() if v == self.dist][0]
        data = {"clusters": b64, "k": self.k, "dist": diststr}
        return data

    @classmethod
    def from_jcr(cls, jcr):
        k = jcr["k"]
        b64 = jcr["clusters"]
        dist = jcr["dist"]
        clusters = np.frombuffer(base64.decodebytes(b64.encode()), dtype=np.float64).reshape((k, -1))
        return cls(k=k, clusters=clusters, dist=dist)



from PIL import ImageFilter
import numpy as np

from rdv.extractors import FeatureExtractor


class Sharpness(FeatureExtractor):
    """Measures the blurryness or sharpness of an iamge. Based on
    https://www.pyimagesearch.com/2015/09/07/blur-detection-with-opencv/
    """

    def __init__(self):
        pass

    def extract_feature(self, data):
        img = data
        img = img.convert("L")
        filtered = img.filter(ImageFilter.Kernel((3, 3), (0, 1, 0, 1, -4, 1, 0, 1, 0), scale=1, offset=0))
        return float(np.array(filtered).mean())  # .var())

    """Serializable inteface """

    def to_jcr(self):
        return {}

    @classmethod
    def from_jcr(cls, jcr):
        return cls()

    """Buildable interface"""

    def build(self, data):
        pass

    def is_built(self):
        return True



import numpy as np
import pandas as pd
from scipy.stats import chisquare, ks_2samp
from abc import ABC, abstractmethod

from rdv.globals import (
    Buildable,
    Serializable,
    DataException,
)


class Stats(Serializable, Buildable, ABC):
    @abstractmethod
    def sample(self, n):
        raise NotImplementedError

    @abstractmethod
    def test_drift(self, other):
        raise NotImplementedError


class NumericStats(Stats):

    _attrs = ["min", "max", "mean", "std", "pinv", "percentiles"]

    def __init__(self, min=None, max=None, mean=None, std=None, pinv=None, percentiles=None):

        self.min = min
        self.max = max
        self.mean = mean
        self.std = std
        self.pinv = pinv
        self.percentiles = percentiles

    """MIN"""

    @property
    def min(self):
        return self._min

    @min.setter
    def min(self, value):
        if value is np.nan:
            raise DataException("stats.min cannot be NaN")
        self._min = value

    """MAX"""

    @property
    def max(self):
        return self._max

    @max.setter
    def max(self, value):
        if value is np.nan:
            raise DataException("stats.max cannot be NaN")
        self._max = value

    """MEAN"""

    @property
    def mean(self):
        return self._mean

    @mean.setter
    def mean(self, value):
        if value is np.nan:
            raise DataException("stats.mean cannot be NaN")
        self._mean = value

    """STD"""

    @property
    def std(self):
        return self._std

    @std.setter
    def std(self, value):
        if value is np.nan:
            raise DataException("stats.std cannot be NaN")
        self._std = value

    """PINV"""

    @property
    def pinv(self):
        return self._pinv

    @pinv.setter
    def pinv(self, value):
        if value is np.nan:
            raise DataException("stats.pinv cannot be NaN")
        self._pinv = value

    """Percentiles"""

    @property
    def percentiles(self):
        return self._percentiles

    @percentiles.setter
    def percentiles(self, value):
        if value is None:
            self._percentiles = None
        elif len(value) == 101:
            self._percentiles = list(value)
        else:
            raise DataException("stats.percentiles must be None or a list of length 101.")

    """Size of the sample that was analyzed"""

    @property
    def samplesize(self):
        return self._samplesize

    @samplesize.setter
    def samplesize(self, value):
        if value is np.nan:
            raise DataException("stats.pinv cannot be NaN")
        self._samplesize = value

    """Serializable Interface"""

    def to_jcr(self):
        data = {}
        for attr in self._attrs:
            data[attr] = getattr(self, attr)
        return data

    @classmethod
    def from_jcr(cls, jcr):
        d = {}
        for attr in cls._attrs:
            d[attr] = jcr[attr]
        return cls(**d)

    """Buildable Interface"""

    def build(self, data):
        data = np.array(data)
        invalids = data[np.isnan(data)]
        data = data[~np.isnan(data)]

        self.min = float(np.min(data))
        self.max = float(np.max(data))
        self.mean = float(data.mean())
        self.std = float(data.std())

        # Build cdf estimate based on percentiles
        q = np.arange(start=0, stop=101, step=1)
        self.percentiles = np.percentile(a=data, q=q)

        # Check the invalid
        self.pinv = len(invalids) / len(data)
        self.samplesize = len(data)

    def is_built(self):
        return all(getattr(self, attr) is not None for attr in self._attrs)

    """Testing and sampling functions"""

    def test_drift(self, other, pthresh=0.05):
        sample_self = self.sample(n=1000)
        sample_other = other.sample(n=1000)
        stat, pvalue = ks_2samp(data1=sample_self, data2=sample_other, alternative="two-sided", mode="auto")
        drift = pvalue < pthresh
        return stat, pvalue, drift

    def sample(self, n=1000, dtype="float"):
        # Sample floats in range 0 - len(percentiles)
        samples = np.random.random(n) * 100

        # We will lineraly interpolate the sample between the percentiles, so get their integer floor and ceil percentile, and the relative diztance from the floor (between 0 and 1)
        floor_percentiles = np.floor(samples).astype("uint8")
        ceil_percentiles = np.ceil(samples).astype("uint8")
        percentiles_alpha = samples - np.floor(samples)

        percentiles = np.array(self.percentiles)
        px = percentiles[floor_percentiles] * (1 - percentiles_alpha) + percentiles[ceil_percentiles] * (
            percentiles_alpha
        )

        if dtype == "int":
            return px.astype(np.int)
        else:
            return px


class CategoricStats(Stats):

    _attrs = ["domain_counts", "pinv", "samplesize"]

    def __init__(self, domain_counts=None, pinv=None, samplesize=None):

        self.domain_counts = domain_counts
        self.pinv = pinv
        self.samplesize = samplesize

    """domain_counts"""

    @property
    def domain_counts(self):
        return self._domain_counts

    @domain_counts.setter
    def domain_counts(self, value):
        if value is None:
            self._domain_counts = value
        elif isinstance(value, dict):
            for key, keyvalue in value.items():
                if keyvalue < 0:
                    raise DataException(f"Domain count for {key} is  < 0")
            self._domain_counts = value
        else:
            raise DataException(f"stats.domain_counts should be a dict, not {type(value)}")

    """PINV"""

    @property
    def pinv(self):
        return self._pinv

    @pinv.setter
    def pinv(self, value):
        if value is np.nan:
            raise DataException("stats.pinv cannot be NaN")
        self._pinv = value

    @property
    def samplesize(self):
        return self._samplesize

    @samplesize.setter
    def samplesize(self, value):
        if value is np.nan:
            raise DataException("stats.pinv cannot be NaN")
        self._samplesize = value

    def to_jcr(self):
        data = {}
        for attr in self._attrs:
            value = getattr(self, attr)
            data[attr] = value
        return data

    @classmethod
    def from_jcr(cls, jcr):
        d = {}
        for attr in cls._attrs:
            d[attr] = jcr[attr]
        return cls(**d)

    def build(self, data):
        data = pd.Series(data)
        self.domain_counts = data[~pd.isna(data)].value_counts(normalize=True).to_dict()
        invalids = data[pd.isna(data)]
        self.pinv = len(invalids) / len(data)
        self.samplesize = len(data)

    def is_built(self):
        return all(getattr(self, attr) is not None for attr in self._attrs)

    """Testing and sampling functions"""

    def test_drift(self, other, pthresh=0.05):
        domain_f1, domain_f2, full_domain = equalize_domains(self.domain_counts, other.domain_counts)
        sample_size = min(self.samplesize, other.samplesize)
        sampled_1 = self.sample_counts(domain_f1, keys=full_domain, n=sample_size)
        sampled_2 = self.sample_counts(domain_f2, keys=full_domain, n=sample_size)

        stat, pvalue = chisquare(f_obs=sampled_1, f_exp=sampled_2)
        drift = pvalue < pthresh
        return stat, pvalue, drift

    def sample(self, n):
        domain = sorted(list(self.domain_counts.keys()))
        # Let's be absolutely sure the domain is always in the same order
        p = [self.domain_counts[k] for k in domain]
        return np.random.choice(a=domain, size=n, p=p)

    def sample_counts(self, domain_freq, keys, n=1000):
        domain = sorted(list(keys))
        # Le's be absolutely sure the domain is always in the same order
        p = [domain_freq.get(k, 0) for k in domain]
        counts = (np.array(p) * (n - len(domain))).astype("int")
        counts += 1  # make sure there are no zeros
        return counts


def add_missing(domain_counts, full_domain):
    for key in full_domain:
        if key not in domain_counts:
            domain_counts[key] = 0
    return domain_counts


def equalize_domains(a, b):
    full_domain = sorted(list(set(set(a.keys()) | set(b.keys()))))
    a = add_missing(a, full_domain)
    b = add_missing(b, full_domain)
    return a, b, full_domain



from jupyter_dash import JupyterDash
import dash


def isnotebook():
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def get_dash(mode="inline"):
    isnb = isnotebook()
    if isnb:
        print("Using JupyterDash")
        return JupyterDash, {"mode": mode}
    else:
        print("Using standard Dash")
        return dash.Dash, {}



#%%
import pytest

import pandas as pd
import numpy as np

import rdv
from rdv.extractors.structured import construct_features
from rdv.feature import CategoricFeature, FloatFeature, IntFeature
from rdv.schema import Schema
from rdv.globals import SchemaStateException, DataException
from rdv.stats import NumericStats, CategoricStats

#%%
def test_constuct_components():
    cols = {
        "num1": list(range(10)),
        "cat1": ["a"] * 5 + ["b"] * 5,
        "cat2": ["c"] * 5 + ["d"] * 5,
        "num2": [0.2] * 10,
    }
    df = pd.DataFrame(data=cols)
    components = construct_features(dtypes=df.dtypes)
    assert len(components) == 4
    assert isinstance(components[0], IntFeature)
    assert isinstance(components[1], CategoricFeature)
    assert isinstance(components[2], CategoricFeature)
    assert isinstance(components[3], FloatFeature)


def test_conmpile():
    cols = {
        "num1": list(range(10)),
        "cat1": ["a"] * 5 + ["b"] * 5,
    }
    df = pd.DataFrame(data=cols)
    components = construct_features(dtypes=df.dtypes)
    schema = Schema(features=components)
    assert not schema.is_built()
    schema.build(data=df)
    components = schema.features
    assert isinstance(components["num1"].stats, NumericStats)
    assert components["num1"].stats.min == 0
    assert components["num1"].stats.max == 9
    assert components["num1"].stats.mean == 4.5
    assert components["num1"].is_built()

    assert isinstance(components["cat1"].stats, CategoricStats)
    assert sorted(components["cat1"].stats.domain_counts.keys()) == sorted(["a", "b"])
    assert components["cat1"].stats.pinv == 0
    assert components["cat1"].is_built()

    assert schema.is_built()


def test_all_nan():
    cols = {
        "num1": [np.nan] * 10,
        "cat1": [np.nan] * 10,
    }
    df = pd.DataFrame(data=cols)
    components = construct_features(dtypes=df.dtypes)
    schema = Schema(features=components)
    try:
        schema.build(data=df)
    except ValueError:
        pass
    else:
        pytest.fail("Feature with all nans should throw a DataException")


# %%



import pytest

import pandas as pd
import numpy as np

import rdv
from rdv.extractors.structured import construct_features
from rdv.feature import CategoricFeature, FloatFeature
from rdv.schema import Schema
from rdv.globals import SchemaStateException, DataException
from rdv.stats import NumericStats, CategoricStats


def test_compile_nan():
    cols = {
        "num1": list(range(10)),
        "cat1": ["a"] * 5 + ["b"] * 5,
    }
    df = pd.DataFrame(data=cols)
    components = construct_features(dtypes=df.dtypes)
    schema = Schema(features=components)
    schema.build(data=df)

    tags = schema.check(pd.Series([np.nan, np.nan], index=["num1", "cat1"]))
    assert len(tags) == 2
    for tag in tags:
        assert tag["type"] == "schema-error"
        assert tag["value"] == "Value NaN"


def test_compile_nan_2():
    cols = {
        "num1": list(range(10)),
        "cat1": ["a"] * 5 + ["b"] * 5,
    }
    df = pd.DataFrame(data=cols)
    components = construct_features(dtypes=df.dtypes)
    schema = Schema(features=components)
    schema.build(data=df)
    components = schema.features

    tags = schema.check(pd.Series([1, "b"], index=["num1", "cat1"]))
    assert len(tags) == 2

    assert tags[0]["type"] == "schema-feature"
    assert tags[0]["name"] == "num1"
    assert tags[0]["value"] == 1
    assert tags[0]["group"] == "default@0.0.0"

    assert tags[1]["type"] == "schema-feature"
    assert tags[1]["name"] == "cat1"
    assert tags[1]["value"] == "b"
    assert tags[1]["group"] == "default@0.0.0"
